# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
# encoding = utf-8
from builtins import object
from . import event_writer
from . import memory_event_writer
from .metric_exception import MetricException
from .number_metric_collector import NumberMetricCollector

from solnlib import pattern
from solnlib import log

import threading
from future.utils import with_metaclass

__all__ = ['Monitor']

logger = log.Logs().get_logger('metric_collector')

class Monitor(with_metaclass(pattern.Singleton, object)):
    EVENT_WRITERS = {
        'memory': memory_event_writer.MemoryEventWriter,
        'file': event_writer.FileEventWriter,
        'hec': event_writer.FileEventWriter  # TODO; implement a HEC writer
    }

    def __init__(self):
        self._app = None
        self._ewriter = None
        self._ewriter_type = None
        self.number_metric_collector = None
        self.worker_start_lock = threading.Lock()

    def configure(self, conf, force_update=False):
        '''
        conf is a dict.
         -- app: the app name
         -- event_writer: the type of event writer.
         -- writer_config: a dict which contains the configuration for the event writer
        force_update: when this is True, update the config of monitor
        '''
        if self.number_metric_collector is not None and force_update is False:
            logger.info('monitor has been initialized.')
            return self

        self._app = conf.get('app', self._app)
        if not self._app:
            raise MetricException('app is not found in configuration object.')

        event_writer = str.lower(conf.get('event_writer', 'file'))

        if self._ewriter_type != event_writer:
            if event_writer not in self.EVENT_WRITERS:
                raise MetricException('Unknown event writer type:' +
                                      event_writer)
            ewriter_cls = self.EVENT_WRITERS[event_writer]
            writer_config = conf.get('writer_config', {})
            self._ewriter_type = event_writer
            self._ewriter = ewriter_cls(self._app, writer_config)
        else:
            writer_config = conf.get('writer_config', {})
            if self._ewriter:
                self._ewriter.update_config(writer_config)
            else:
                raise MetricException('event writer type is ' +
                                      self._ewriter_type +
                                      ', while event writer is None.')

        if self.number_metric_collector is None:
            self.number_metric_collector = NumberMetricCollector(self._ewriter)
        else:
            self.number_metric_collector.set_event_writer(self._ewriter)
        logger.info('Initialize monitor successfully. app=%s, config:%s', self._app, conf)
        return self

    def start(self):
        with self.worker_start_lock:
            if self.number_metric_collector:
                if self.number_metric_collector.is_stopped():
                    self.number_metric_collector.start()
                else:
                    logger.info('Collector thread has been started.')
            else:
                raise ValueError('Monitor is not configued yet.')

    def stop(self):
        if self.number_metric_collector:
            self.number_metric_collector.stop()
            self.number_metric_collector.join()

    def is_stopped(self):
        if self.number_metric_collector:
            return self.number_metric_collector.is_stopped()
        else:
            return True

    def write_event(self, ev, tags=[]):
        if self._ewriter:
            self._ewriter.write_event(ev, tags)

    def register_metric(self,
                        metric_name,
                        metric_type=NumberMetricCollector.SUM_METRIC,
                        metric_tags=[],
                        max_time_span=10):
        self.number_metric_collector.register_metric(
            metric_name, metric_type, metric_tags, max_time_span)

    def record_metric(self, metric_name, metric_value, metric_timestamp=None):
        if self.number_metric_collector:
            self.number_metric_collector.record_metric(metric_name, metric_value, metric_timestamp)



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import print_function
import shutil
import errno
import json
import os
from jinja2 import Environment, FileSystemLoader

from uccrestbuilder.global_config import GlobalConfigBuilderSchema, GlobalConfigPostProcessor
from uccrestbuilder import build


basedir = os.path.dirname(os.path.abspath(__file__))
top_dir = os.path.dirname(basedir)

# jinja2 environment
j2_env = Environment(loader=FileSystemLoader(top_dir))

# read schema from globalConfig.json
root_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

with open(os.path.join(root_path, 'globalConfig.json')) as f:
    json_schema = ''.join([l for l in f])

schema_content = json.loads(json_schema)
scheme = GlobalConfigBuilderSchema(schema_content)
ta_name = schema_content.get("meta").get("name")
ta_namespace = schema_content.get("meta").get("restRoot")
import_declare_name = 'import_declare_test'


def clean_before_build():
    shutil.rmtree(os.path.join(basedir, 'build'), ignore_errors=True)


def generate_rest():
    build(
        scheme,
        'splunktaucclib.rest_handler.admin_external.AdminExternalHandler',
        os.path.join('.', 'build'),
        post_process=GlobalConfigPostProcessor(),
        import_declare_name=import_declare_name
    )


def copy_directory(src, dest):
    try:
        shutil.copytree(src, dest)
    except OSError as exc:
        if exc.errno == errno.ENOTDIR:
            shutil.copy(src, dest)
        else:
            print('Directory %s not copied. Error: %s' % (src, exc))


def add_modular_input():
    services = schema_content.get("pages").get("inputs").get("services")
    for service in services:
        input_name = service.get("name")
        class_name = input_name.upper()
        description = service.get("title")
        entity = service.get("entity")
        field_allow_list = ["name", "index", "sourcetype"]
        template = "input.template"
        # if the service has a template specified, use it.  Otherwise keep the default
        if "template" in service:
            template = service.get("template") + ".template"

        # filter fields in allow list
        entity = [x for x in entity if x.get("field") not in field_allow_list]
        import_declare = 'import ' + import_declare_name

        content = j2_env.get_template(os.path.join('templates', template)).render(
            import_declare=import_declare,
            input_name=input_name,
            class_name=class_name,
            description=description,
            entity=entity
        )
        input_file_name = os.path.join(
            basedir,
            'build',
            'bin',
            input_name + '.py'
        )
        with open(input_file_name, "w") as input_file:
            input_file.write(content)


clean_before_build()
generate_rest()
add_modular_input()



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
import sys
import os.path as op
from os import rename, remove, makedirs
from .alert_actions_merge import merge_conf_file

from splunk_add_on_ucc_framework.alert_utils.alert_utils_common.conf_parser import TABConfigParser


def write_file(file_name, file_path, content, logger, merge="stanza_overwrite"):
    logger.debug('operation="write", object="%s" object_type="file"',
                 file_path)

    do_merge = False
    if file_name.endswith('.conf') or file_name.endswith('conf.spec'):
        do_merge = True
    else:
        logger.info('event="Will not merge file="%s", ' +
                    'reason="Only support conf file merge"', file_path)

    if file_path:
        new_file = None
        if op.exists(file_path) and do_merge:
            new_file = op.join(op.dirname(file_path), "new_" + file_name)
        if new_file:
            try:
                with open(new_file, 'w+') as fhandler:
                    fhandler.write(content)
                merge_conf_file(new_file, file_path, merge)
            finally:
                if op.exists(new_file):
                    remove(new_file)
        else:
            if not op.exists(op.dirname(file_path)):
                makedirs(op.dirname(file_path))
            with open(file_path, 'w+') as fhandler:
                fhandler.write(content)
            if do_merge:
                # need to process the file with conf parser
                parser = TABConfigParser()
                parser.read(file_path)
                with open(file_path, 'w') as df:
                    parser.write(df)
    else:
        sys.stdout.write("\n##################File {}##################\n".format(file_name))
        sys.stdout.write(content)


GLOBAL_SETTING_TYPE_MAP = {
    "text": "text",
    "checkbox": "bool",
    "password": "password"
}

GLOBAL_SETTING_VALUE_NAME_MAP = {
    "text": "content",
    "bool": "bool",
    "password": "password"
}


def convert_custom_setting(parameters):
    """
    convert
    [{
        "default_value": "message",
        "name": "notification_type",
        "required": true,
        "help_string": "Choose style of HipChat notification.",
        "possible_values": {
             "Application Card": "card",
             "Message": "message"
         },
         "label": "Notification Style",
         "format_type": "dropdownlist",
         "value": "xxxx"
      }]
    to
    [{
    "title": "customized key",
    "name": "customized name",
    "type": "text",
    "description": "description of customized key"
    }]
    """
    formated = []
    if not parameters:
        return formated

    for param in parameters:
        if param.get("format_type") not in list(GLOBAL_SETTING_TYPE_MAP.keys()):
            msg = 'format_type="{}" is not support for global setting'.format(
                param.get("format_type"))
            raise Exception(msg)

        one_param = {
            "title": param.get("label"),
            "name": param.get("name"),
            "type": GLOBAL_SETTING_TYPE_MAP[param.get("format_type")],
            "description": param.get("help_string")
        }
        formated.append(one_param)

    return formated


def convert_global_setting(global_settings):
    """
    convert
    {
    "customized_settings": {
        "string_label": {
        "type": "text",
        "content": "string"
        },
        "password": {
        "type": "password",
        "password": "123"
        },
        "checkbox": {
        "type": "bool",
        "bool": true
        }
    },
    "proxy_settings": {
        "proxy_password": "sef",
        "proxy_type": "http",
        "proxy_url": "1.2.3.4",
    },
    "global_settings": {
        "log_level": "INFO"
    }
  to
    {
    "title": "Proxy",
    "name": "proxy",
    "type": "default_proxy",
    "description": "proxy settings"
    },
    {
    "title": "Account Key Title",
    "name": "username",
    "type": "default_account",
    "description": "The username of the user account"
    },
    {
    "title": "Account Secret Title",
    "name": "password",
    "type": "default_account",
    "description": "The password of the user account"
    },
    {
    "title": "customized key",
    "name": "customized name",
    "type": "text",
    "description": "description of customized key"
    """
    converted = []
    if not global_settings:
        return converted

    for type, settings in list(global_settings.items()):
        if type == "proxy_settings":
            proxy = {
                "title": "Proxy",
                "name": "proxy",
                "type": "default_proxy",
                "description": "proxy settings"}
            converted.append(proxy)
        elif type == "log_settings":
            logging = {
                "title": "Logging",
                "name": "logging",
                "type": "default_logging",
                "description": "logging setting"}
            converted.append(logging)
        elif type == "credential_settings":
            username = {
                "title": "Account Key Title",
                "name": "tab_default_account_username",
                "type": "default_account",
                "description": "The username of the user account"
            }
            password = {
                "title": "Account Secret Title",
                "name": "tab_default_account_password",
                "type": "default_account",
                "description": "The password of the user account"}
            converted.append(username)
            converted.append(password)
        elif type == "customized_settings":
            custom_settings = convert_custom_setting(settings)
            converted += custom_settings
    return converted



def convert_global_setting_previous(global_settings):
    """
    convert global_settings=[
        {
            "type": "proxy"
        },
        {
            "type": "logging"
        },
        {
            "type": "account"
        },
        {
            "type": "custom",
            "parameters": []
        }
    ]
        to [
    {
    "title": "Proxy",
    "name": "proxy",
    "type": "default_proxy",
    "description": "proxy settings"
    },
    {
    "title": "Account Key Title",
    "name": "username",
    "type": "default_account",
    "description": "The username of the user account"
    },
    {
    "title": "Account Secret Title",
    "name": "password",
    "type": "default_account",
    "description": "The password of the user account"
    },
    {
    "title": "customized key",
    "name": "customized name",
    "type": "text",
    "description": "description of customized key"
    }
]
    """
    converted = []
    if not global_settings:
        return converted

    for setting in global_settings:
        if setting.get("type") == "proxy":
            proxy = {
                "title": "Proxy",
                "name": "proxy",
                "type": "default_proxy",
                "description": "proxy settings"}
            converted.append(proxy)
        elif setting.get("type") == "logging":
            logging = {
                "title": "Logging",
                "name": "logging",
                "type": "default_logging",
                "description": "logging setting"}
            converted.append(logging)
        elif setting.get("type") == "account":
            username = {
                "title": "Account Key Title",
                "name": "username",
                "type": "default_account",
                "description": "The username of the user account"
            }
            password = {
                "title": "Account Secret Title",
                "name": "password",
                "type": "default_account",
                "description": "The password of the user account"}
            converted.append(username)
            converted.append(password)
        elif setting.get("type") == "custom":
            custom_settings = convert_custom_setting(setting.get("parameters"))
            converted += custom_settings
    return converted


def get_test_parameter_type(param):
    if not param:
        return None

    if param.get("format_type") in list(GLOBAL_SETTING_TYPE_MAP.keys()):
        return GLOBAL_SETTING_TYPE_MAP.get(param.get("format_type"))
    return None


def get_parameter_type(param, parameters_meta, logger):
    if not parameters_meta:
        logger.info('parameters_meta="None"')
        return None

    if not param:
        logger.info('param="None"')
        return None

    for param_meta in parameters_meta:
        if param == param_meta["name"]:
            return GLOBAL_SETTING_TYPE_MAP.get(param_meta["format_type"])

    return None


def convert_test_global_settings(test_global_settings, logger):
    """
    convert to:
{
  "customized_settings": {
    "string_label": {
      "type": "text",
      "content": "string"
    },
    "password": {
      "type": "password",
      "password": "123"
    },
    "checkbox": {
      "type": "bool",
      "bool": true
    }
  },
  "proxy_settings": {
    "proxy_password": "sef",
    "proxy_type": "http",
    "proxy_url": "1.2.3.4",
    "proxy_rdns": "0",
    "proxy_username": "sdf",
    "proxy_port": "34",
    "proxy_enabled": "1"
  },
  "global_settings": {
    "log_level": "INFO"
  }
}
    """
    if not test_global_settings:
        logger.info('test_global_settings="%s"', test_global_settings)
        return {}

    converted = {}
    for type, settings in list(test_global_settings.items()):
        if type == "customized_settings":
            converted["customized_settings"] = {}
            for setting in settings:
                type = get_test_parameter_type(setting)
                if not type:
                    msg = 'No type for {} in customized_settings'.format(setting)
                    raise NotImplementedError(msg)

                converted["customized_settings"][setting["name"]] = {
                    "type": type,
                    GLOBAL_SETTING_VALUE_NAME_MAP[type]: setting.get("value")
                }
        elif type == "log_settings":
            converted["global_settings"] = settings
        else:
            converted[type] = settings
    return converted


def split_path(path):
    """
    split a path into a list
    """
    if not path:
        return None
    paths = []
    (head, tail) = op.split(path)
    while tail:
        paths.insert(0, tail)
        (head, tail) = op.split(head)
    (drive, rest) = op.splitdrive(head)
    if drive:
        paths.insert(0, drive)
    else:
        paths.insert(0, head)
    return paths



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from . import normalize
import logging
import os
from .modular_alert_builder.build_core import generate_alerts
import traceback


class LoggerAdapter(logging.LoggerAdapter):
    def __init__(self, prefix, logger):
        super(LoggerAdapter, self).__init__(logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        return '[%s] %s' % (self.prefix, msg), kwargs


def validate(alert, logger):
    try:
        fields = []
        if alert.get("entity"):
            for entity in alert.get("entity"):
                if entity.get("field") in fields:
                    raise Exception("Field names should be unique")
                else:
                    fields.append(entity.get("field"))

                if entity.get("type") in ["radio", "singleSelect"]:
                    if not entity.get("options"):
                        raise Exception(
                            "{} type must have options parameter".format(entity.get("type")))
                elif entity.get("options"):
                    raise Exception(
                        "{} type must not contain options parameter".format(entity.get("type")))

                if entity.get("type") in ["singleSelectSplunkSearch"]:
                    if not all([entity.get("search"), entity.get("valueField"), entity.get("labelField")]):
                        raise Exception("{} type must have search, valueLabel and valueField parameters".format(
                            entity.get("type")))
                elif any([entity.get("search"), entity.get("valueField"), entity.get("labelField")]):
                    raise Exception("{} type must not contain search, valueField or labelField parameter".format(
                        entity.get("type")))

    except:
        logger.error(traceback.format_exc())
        raise


def alert_build(schema_content, product_id, short_name, output_dir,sourcedir):

    # Initializing logger
    logging.basicConfig()
    logger = logging.getLogger('Alert Logger')
    logger = LoggerAdapter('ta="{}" Creating Alerts'.format(short_name),
                           logger)

    # Validation
    for alert in schema_content['alerts']:
        validate(alert, logger)

    # Get the alert schema with required structure
    envs = normalize.normalize(schema_content, product_id, short_name)
    pack_folder = os.path.join(sourcedir, 'arf_dir_templates', 'modular_alert_package')
    # Generate Alerts
    generate_alerts(pack_folder, output_dir, logger, envs)



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import object
import threading
import queue
import time

from solnlib import log
from .metric_exception import MetricException
from . import metric_aggregator

__all__ = ['NumberMetricCollector']

logger = log.Logs().get_logger('metric_collector')

class AggregatorWorker(threading.Thread):

    def __init__(self, collector):
        super(AggregatorWorker, self).__init__(name='AggregatorWorker')
        self.daemon = True
        self.collector = collector
        self.metric_aggregators = dict()

    def run(self):
        logger.info("Start aggregator worker.")
        stop = False
        while not stop:
            stop = self.collector.process()
        self.collector.flush_all_event()


class NumberMetricCollector(object):
    SUM_METRIC = 'sum'
    AVG_METRIC = 'avg'

    def __init__(self, event_writer):
        self.event_writer = event_writer
        self.record_queue = queue.Queue()
        self.aggregators = dict()
        self.worker = None

    def set_event_writer(self, ew):
        self.event_writer = ew

    def is_stopped(self):
        if self.worker is None:
            return True
        else:
            return not self.worker.is_alive()

    def register_metric(self,
                        metric_name,
                        metric_type=SUM_METRIC,
                        metric_tags=[],
                        max_time_span=10):
        if metric_name in self.aggregators:
            raise MetricException('metric {} has been registered.'.format(
                metric_name))
        if metric_type == self.AVG_METRIC:
            self.aggregators[
                metric_name] = metric_aggregator.NumberMetricAverage(
                    metric_name, self.event_writer, metric_tags, max_time_span)
        elif metric_type == self.SUM_METRIC:
            self.aggregators[metric_name] = metric_aggregator.NumberMetricSum(
                metric_name, self.event_writer, metric_tags, max_time_span)
        else:
            raise MetricException('Metric type {} is unsupported.'.format(
                metric_type))

    def record_metric(self, metric_name, metric_value, metric_timestamp=None):
        if metric_name not in self.aggregators:
            raise MetricException('metric {} is not registered yet.'.format(
                metric_name))
        if metric_timestamp is None:
            metric_timestamp = int(time.time())
        self.record_queue.put({'type': 'metric',
                               'ts': metric_timestamp,
                               'v': metric_value,
                               'n': metric_name})

    def start(self):
        if self.is_stopped() == False:
            raise RuntimeError('collector worker has been started.')
        if self.worker:
            raise RuntimeError('Worker thread is stopped, but the worker is not None.')
        self.worker = AggregatorWorker(self)
        self.worker.start()

    def stop(self, graceful=True):
        self.record_queue.put({'type': 'stop'})
        if graceful:
            self.join()
        self.worker = None

    def join(self):
        if self.worker:
            self.worker.join()

    def process(self):
        '''
        the entry point for the worker thread
        '''
        is_stop = False
        skip_metrics = []
        record = None
        try:
            record = self.record_queue.get(timeout=1)
        except queue.Empty:
            pass
        if record:
            if record['type'] == 'stop':
                is_stop = True
            else:
                metric_name = record['n']
                self.aggregators[metric_name].aggregate_metric(record)
                skip_metrics.append(metric_name)
        self._flush_events(skip_metrics)
        return is_stop

    def _flush_events(self, skip_metrics=[]):
        ts = int(time.time())
        for k, aggr in list(self.aggregators.items()):
            if k in skip_metrics:
                continue
            aggr._flush_buckets(ts)

    def flush_all_event(self):
        for k, v in list(self.aggregators.items()):
            v.flush_all_event()



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import os.path as op


class AlertActionsTemplateMgr(object):
    CURRENT_DIR = op.dirname(op.abspath(__file__))
    DEFAULT_TEMPLATE_DIR = op.join(CURRENT_DIR, "arf_template")
    DEFAULT_HTML_LOOKUP_DIR = op.join(DEFAULT_TEMPLATE_DIR, "default_html_theme")

    def __init__(self, template_dir=None, html_theme=None):
        self._template_dir = template_dir or \
            AlertActionsTemplateMgr.DEFAULT_TEMPLATE_DIR
        self._html_theme = html_theme or \
            AlertActionsTemplateMgr.DEFAULT_HTML_LOOKUP_DIR

    def get_template_dir(self):
        return self._template_dir

    def get_html_lookup_dir(self):
        return self._html_theme



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

# encode=utf-8

class MetricException(Exception):
    def __init__(self, *args, **kwargs):
        super(MetricException, self).__init__(*args, **kwargs)



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import

from .base import indent
from .single_model import RestEndpointBuilder, RestEntityBuilder


class MultipleModelEntityBuilder(RestEntityBuilder):
    @property
    def name_spec(self):
        return self.name

    @property
    def name_default(self):
        return self.name

    @property
    def name_rh(self):
        return "_" + self._name


class MultipleModelEndpointBuilder(RestEndpointBuilder):

    _rh_template = """
from splunktaucclib.rest_handler.endpoint import (
    field,
    validator,
    RestModel,
    MultipleModel,
)
from splunktaucclib.rest_handler import admin_external, util
from {handler_module} import {handler_name}
import logging

util.remove_http_proxy_env_vars()

{entities}

endpoint = MultipleModel(
    '{conf_name}',
    models=[
{models}
    ],
)


if __name__ == '__main__':
    logging.getLogger().addHandler(logging.NullHandler())
    admin_external.handle(
        endpoint,
        handler={handler_name},
    )
"""

    def actions(self):
        return ["edit", "list"]

    def generate_rh(self, handler):
        entities = [entity.generate_rh() for entity in self._entities]
        models = ["model" + entity.name_rh for entity in self._entities]
        models_lines = ", \n".join(models)
        return self._rh_template.format(
            handler_module=handler.module,
            handler_name=handler.name,
            entities="\n".join(entities),
            models=indent(models_lines, 2),
            conf_name=self.conf_name,
        )



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

__version__ = "4.4.0"

import logging
import os, time
import re
import glob
from os import system
import shutil
import sys
import argparse
import json
from xml.etree import cElementTree as et
from .uccrestbuilder.global_config import (
    GlobalConfigBuilderSchema,
    GlobalConfigPostProcessor,
)
from .uccrestbuilder import build
from .start_alert_build import alert_build

import jsonschema
from jinja2 import Environment, FileSystemLoader
from dunamai import Version, Style
import configparser

outputdir = os.path.join(os.getcwd(), "output")
sourcedir = os.path.dirname(os.path.realpath(__file__))

j2_env = Environment(
    loader=FileSystemLoader(os.path.join(sourcedir, "templates"))
)

logger = logging.getLogger('UCC')
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s [%(name)s] %(levelname)s: %(message)s')
shandler = logging.StreamHandler()
shandler.setLevel(logging.INFO)
shandler.setFormatter(formatter)
logger.addHandler(shandler)

PARENT_DIR = ".."


def get_os_path(path):
    """
    Returns a path which will be os compatible.

    Args:
        path (str): Path in string

    Return:
        string: Path which will be os compatible.
    """

    if "\\\\" in path:
        path = path.replace("\\\\", os.sep)
    else:
        path = path.replace("\\", os.sep)
    path = path.replace("/", os.sep)
    return path.strip(os.sep)

def recursive_overwrite(src, dest, ignore_list=None):
    """
    Method to copy from src to dest recursively.

    Args:
        src (str): Source of copy
        dest (str): Destination to copy
        ignore_list (list): List of files/folder to ignore while copying
    """

    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        for f in files:
            if not ignore_list or not os.path.join(dest, f) in ignore_list:
                recursive_overwrite(
                    os.path.join(src, f), os.path.join(dest, f), ignore_list
                )
            else:
                logger.info("Excluding : {}".format(os.path.join(dest, f)))
    else:
        if os.path.exists(dest):
            os.remove(dest)
        shutil.copy(src, dest)


def clean_before_build():
    """
    Clean output directory before build process.
    """

    logger.info("Cleaning out directory " + outputdir)
    shutil.rmtree(os.path.join(outputdir), ignore_errors=True)
    os.makedirs(os.path.join(outputdir))
    logger.info("Cleaned out directory " + outputdir)

def version_tuple(version_str):
    """
    convert string into tuple to compare version

    Args:
        version_str : raw string
    Returns:
        tuple : version into tupleformat
    """
    filled = []
    for point in version_str.split("."):
        filled.append(point.zfill(8))
    return tuple(filled)

def handle_update(config_path):
    """
    handle changes in globalConfig.json

    Args:
        config_path : path to globalConfig.json

    Returns:
        dictionary : schema_content (globalConfig.json)     
    """
    with open(config_path, "r") as config_file:
        schema_content = json.load(config_file)
    # check for schemaVersion in meta, if not availble then set default 0.0.0 
    version = schema_content.get("meta").get("schemaVersion","0.0.0")

    # check for schemaVersion, if it's less than 0.0.1 then updating globalConfig.json 
    if version_tuple(version) < version_tuple("0.0.1"):
        ta_tabs = schema_content.get("pages").get("configuration",{}).get("tabs",{})

        # check in every Account tab for biased term
        for tab in ta_tabs:
            conf_entitties= tab.get("entity")
            for entity in conf_entitties:
                entity_option = entity.get("options")
                if entity_option and "whiteList" in entity_option:
                    entity_option["allowList"] = entity_option.get("whiteList")
                    del entity_option["whiteList"]
                if entity_option and "blackList" in entity_option:
                    entity_option["denyList"] = entity_option.get("blackList")
                    del entity_option["blackList"]

        is_inputs = ("inputs" in schema_content.get("pages"))
        if is_inputs:
            services = schema_content.get("pages").get("inputs",{}).get("services",{})
            # check in every Input service for biased term
            for service in services:
                conf_entitties= service.get("entity")
                for entity in conf_entitties:
                    entity_option = entity.get("options")
                    if entity_option and "whiteList" in entity_option:
                        entity_option["allowList"] = entity_option.get("whiteList")
                        del entity_option["whiteList"]
                    if entity_option and "blackList" in entity_option:
                        entity_option["denyList"] = entity_option.get("blackList")
                        del entity_option["blackList"]

        # set schemaVersion to 0.0.1 as updated globalConfig.json according to new update
        schema_content["meta"]["schemaVersion"]="0.0.1"

        # upadating new changes in globalConfig.json 
        with open(config_path, "w") as config_file:
            json.dump(schema_content,config_file, ensure_ascii=False, indent=4)
    return schema_content
    

def copy_package_source(args, ta_name):
    """
    Copy source package to output directory.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
        ta_name (str): Name of TA.
    """

    logger.info("Copy package directory ")
    recursive_overwrite(args.source, os.path.join(outputdir, ta_name))


def replace_token(args, ta_name):
    """
    Replace token with addon name in inputs.xml, configuration.xml, redirect.xml.
    Replace token with addon version in redirect.xml.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
        ta_name (str): Name of TA.
    """

    # replace token in template
    logger.info("Replace tokens in views")
    views = ["inputs.xml", "configuration.xml", "redirect.xml"]
    for view in views:
        template_dir = os.path.join(
            outputdir, ta_name, "default", "data", "ui", "views"
        )
        with open(os.path.join(template_dir, view)) as f:
            s = f.read()

        # Safely write the changed content, if found in the file
        with open(os.path.join(template_dir, view), "w") as f:
            s = s.replace("${package.name}", ta_name)
            if view == "redirect.xml":
                s = s.replace("${ta.name}", ta_name.lower())
            f.write(s)

def install_libs(path, ucc_lib_target):
    """
    Install 3rd Party libraries in addon.

    Args:
        parent_path (str): Path of parent directory.
        ucc_lib_target (str): Target path to install libraries.
    """

    def _install_libs(requirements, ucc_target, installer="python3"):
        """
        Install 3rd Party libraries using pip2/pip3

        Args:
            requirements (str): Path to requirements file.
            ucc_target (str): Target path to install libraries.
            installer (str): Pip version(pip2/pip3).
        """
        if not os.path.exists(requirements):
            logging.warning("Unable to find requirements file. {}".format(requirements))
        else:
            if not os.path.exists(ucc_target):
                os.makedirs(ucc_target)
            install_cmd = (
                installer +" -m pip install -r \""
                + requirements
                + "\" --no-compile --prefer-binary --ignore-installed --use-deprecated=legacy-resolver --target \""
                + ucc_target
                + "\""
            )
            os.system(installer +" -m pip install pip --upgrade")
            os.system(install_cmd)
            remove_files(ucc_target)
    logging.info(f"  Checking for requirements in {path}")
    if os.path.exists(os.path.join(path,"lib", "requirements.txt")):
        logging.info(f"  Uses common requirements")    
        _install_libs(requirements=os.path.join(path, "lib","requirements.txt"), ucc_target=ucc_lib_target)
    elif os.path.exists(os.path.join(os.path.abspath(os.path.join(path, os.pardir)), "requirements.txt")):
        logging.info(f"  Uses common requirements")    
        _install_libs(requirements=os.path.join(os.path.abspath(os.path.join(path, os.pardir)), "requirements.txt"), ucc_target=ucc_lib_target)
    else:
        logging.info(f"  Not using common requirements")    

    if os.path.exists(os.path.join(path,"lib","py2", "requirements.txt")):
        logging.info(f"  Uses py2 requirements")    
        _install_libs(requirements=os.path.join(path,"lib","py2", "requirements.txt"), installer="python2", ucc_target=os.path.join(ucc_lib_target, "py2"))
    elif os.path.exists(os.path.join(os.path.abspath(os.path.join(path, os.pardir)), "requirements_py2.txt")):
        logging.info(f"  Uses py2 requirements")    
        _install_libs(requirements=os.path.join(os.path.abspath(os.path.join(path, os.pardir)), "requirements_py2.txt"), installer="python2", ucc_target=os.path.join(ucc_lib_target, "py2"))        
    else:
        logging.info(f"  Not using py2 requirements")    

    if os.path.exists(os.path.join(path, "lib","py3","requirements.txt")):
        logging.info(f"  Uses py3 requirements")            
        _install_libs(requirements=os.path.join(path,"lib", "py3","requirements.txt"), ucc_target=os.path.join(ucc_lib_target, "py3"))
    elif os.path.exists(os.path.join(os.path.abspath(os.path.join(path, os.pardir)), "requirements_py3.txt")):
        logging.info(f"  Uses py3 requirements")    
        _install_libs(requirements=os.path.join(os.path.abspath(os.path.join(path, os.pardir)), "requirements_py3.txt"), installer="python3", ucc_target=os.path.join(ucc_lib_target, "py2"))        
    else:
        logging.info(f"  Not using py3 requirements")    


def remove_files(path):
    """
    Remove *.egg-info and *.dist-info files in given path.

    Args:
        path (str): Path to remove *.egg-info and *.dist-info files.
    """

    rmdirs = glob.glob(os.path.join(path, "*.egg-info")) + glob.glob(os.path.join(path, "*.dist-info"))
    for rmdir in rmdirs:
        shutil.rmtree(rmdir)

def generate_rest(args, ta_name, scheme, import_declare_name):
    """
    Build REST for Add-on.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
        ta_name (str): Name of TA.
        scheme (GlobalConfigBuilderSchema): REST schema.
        import_declare_name (str): Name of import_declare_* file.
    """

    build(
        scheme,
        "splunktaucclib.rest_handler.admin_external.AdminExternalHandler",
        os.path.join(outputdir, ta_name),
        j2_env,
        post_process=GlobalConfigPostProcessor(),
        import_declare_name=import_declare_name,
    )


def is_oauth_configured(ta_tabs):
    """
    Check if oauth is configured in globalConfig.json.

    Args:
        ta_tabs (list): List of tabs mentioned in globalConfig.json.

    Returns:
        bool: True if oauth is configured, False otherwise.
    """

    for tab in ta_tabs:
        if tab["name"] == "account":
            for elements in tab["entity"]:
                if elements["type"] == "oauth":
                    return True
            break
    return False


def replace_oauth_html_template_token(args, ta_name, ta_version):
    """
    Replace tokens with addon name and version in redirect.html.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
        ta_name (str): Name of TA.
        ta_version (str): Version of TA.
    """

    html_template_path = os.path.join(
        outputdir, ta_name, "appserver", "templates"
    )
    with open(os.path.join(html_template_path, "redirect.html")) as f:
        s = f.read()

    # Safely write the changed content, if found in the file
    with open(os.path.join(html_template_path, "redirect.html"), "w") as f:
        # replace addon name in html template
        s = s.replace("${ta.name}", ta_name.lower())
        # replace addon version in html template
        s = s.replace("${ta.version}", ta_version)
        f.write(s)


def modify_and_replace_token_for_oauth_templates(
    args, ta_name, ta_tabs, ta_version
):
    """
    Rename templates with respect to addon name if OAuth is configured.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
        ta_name (str): Name of TA.
        ta_version (str): Version of TA.
        ta_tabs (list): List of tabs mentioned in globalConfig.json.

    """
    redirect_xml_src = os.path.join(
        outputdir, ta_name, "default", "data", "ui", "views", "redirect.xml"
    )
    redirect_js_src = os.path.join(
        outputdir, ta_name, "appserver", "static", "js", "build", "redirect_page.js"
    )
    redirect_html_src = os.path.join(
        outputdir, ta_name, "appserver", "templates", "redirect.html"
    )

    if is_oauth_configured(ta_tabs):
        replace_oauth_html_template_token(args, ta_name, ta_version)

        redirect_js_dest = (
            os.path.join(outputdir, ta_name, "appserver", "static", "js", "build", "")
            + ta_name.lower()
            + "_redirect_page."
            + ta_version
            + ".js"
        )
        redirect_html_dest = (
            os.path.join(outputdir, ta_name, "appserver", "templates", ta_name.lower() + "_redirect.html")
        )
        redirect_xml_dest = (
            os.path.join(outputdir, ta_name, "default", "data", "ui", "views", ta_name.lower() + "_redirect.xml")
        )
        os.rename(redirect_js_src, redirect_js_dest)
        os.rename(redirect_html_src, redirect_html_dest)
        os.rename(redirect_xml_src, redirect_xml_dest)

    # if oauth is not configured remove the extra template
    else:
        os.remove(redirect_xml_src)
        os.remove(redirect_html_src)
        os.remove(redirect_js_src)

def add_modular_input(
    args, ta_name, schema_content, import_declare_name
):
    """
    Generate Modular input for addon.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
        ta_name (str): Name of TA.
        schema_content (dict): JSON schema of globalConfig.json
    """

    services = schema_content.get("pages").get("inputs").get("services")
    for service in services:
        input_name = service.get("name")
        class_name = input_name.upper()
        description = service.get("title")
        entity = service.get("entity")
        field_allow_list = ["name", "index", "sourcetype"]
        template = "input.template"
        # if the service has a template specified, use it.  Otherwise keep the default
        if "template" in service:
            template = service.get("template") + ".template"

        # filter fields in allow list
        entity = [x for x in entity if x.get("field") not in field_allow_list]
        import_declare = "import " + import_declare_name

        content = j2_env.get_template(template).render(
            import_declare=import_declare,
            input_name=input_name,
            class_name=class_name,
            description=description,
            entity=entity,
        )
        input_file_name = os.path.join(
            outputdir, ta_name, "bin", input_name + ".py"
        )
        with open(input_file_name, "w") as input_file:
            input_file.write(content)

        input_default = os.path.join(
            outputdir, ta_name, "default",  "inputs.conf"
        )   
        config = configparser.ConfigParser()
        if os.path.exists(input_default):
            config.read(input_default)
        
        if config.has_section(input_name):
            config[input_name]['python.version'] = 'python3'
        else:
            config[input_name] = {'python.version': 'python3'}
        
        with open(input_default, 'w') as configfile:
           config.write(configfile)


def make_modular_alerts(args, ta_name, ta_namespace, schema_content):
    """
    Generate the alert schema with required structure.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
        ta_name (str): Name of TA.
        ta_namespace (str): restRoot of TA.
        schema_content (dict): JSON schema of globalConfig.json.

    """

    if schema_content.get("alerts"):
        alert_build(
            {"alerts": schema_content["alerts"]},
            ta_name,
            ta_namespace,
            outputdir,
            sourcedir,
        )

def get_ignore_list(ta_name, path):
    """
    Return path of files/folders to be removed.

    Args:
        ta_name (str): Name of TA.
        path (str): Path of '.uccignore'.

    Returns:
        list: List of paths to be removed from output directory.
    """
    if not os.path.exists(path):
        return []
    else:
        with open(path) as ignore_file:
            ignore_list = ignore_file.readlines()
        ignore_list = [(os.path.join("output", ta_name, get_os_path(path))).strip() for path in ignore_list]
        return ignore_list

def remove_listed_files(ignore_list):
    """
    Return path of files/folders to removed in output folder.

    Args:
        ignore_list (list): List of files/folder to removed in output directory.

    """
    for path in ignore_list:
        if os.path.exists(path):
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path, ignore_errors=True)
        else:
            logger.info("While ignoring the files mentioned in .uccignore {} was not found".format(path))

def update_ta_version(config, ta_version):
    """
    Update version of TA in globalConfig.json.

    Args:
        args (argparse.Namespace): Object with command-line arguments.
    """

    with open(config, "r") as config_file:
        schema_content = json.load(config_file)
    schema_content.setdefault("meta", {})["version"] = ta_version
    with open(config, "w") as config_file:
        json.dump(schema_content, config_file, indent=4)

def handle_no_inputs(ta_name):
    """
    Handle for configuration without input page.

    Args:
        ta_name (str): Name of TA. 
    """
    def _removeinput(path):
        """
        Remove "inputs" view from default.xml

        Args:
            path (str) : path to default.xml
        """
        tree = et.parse(path)
        root = tree.getroot()

        for element in root:
            if element.tag =="view" and element.get('name') == "inputs":
                root.remove(element)

        tree.write(path)

    default_xml_file = os.path.join(
        outputdir, ta_name, "default", "data", "ui", "nav","default.xml"
    )
    # Remove "inputs" view from default.xml
    _removeinput(default_xml_file)

    file_remove_list = []
    file_remove_list.append(os.path.join(
        outputdir, ta_name, "default", "data", "ui", "views","inputs.xml"
    ))
    file_remove_list.append(os.path.join(outputdir,ta_name,"appserver","static","css","inputs.css"))
    file_remove_list.append(os.path.join(outputdir,ta_name,"appserver","static","css","createInput.css"))
    # Remove unnecessary files
    for fl in file_remove_list:
        try:
            os.remove(fl)
        except OSError:
            pass
    
def save_comments(outputdir, ta_name):
    """
    Save index and content of comments in conf file and return dictionary thereof
    """
    config_file = os.path.join(outputdir, ta_name,'default', "app.conf")
    comment_map = {}
    with open(config_file, 'r') as file:
        i = 0
        lines = file.readlines()
        for line in lines:
            if re.match( r'^\s*#.*?$', line):
                comment_map[i] = line
            i += 1
    return comment_map

def restore_comments(outputdir, ta_name, comment_map):
    """
    Write comments to conf file at their original indices
    """
    config_file = os.path.join(outputdir, ta_name,'default', "app.conf")
    with open(config_file, 'r') as file:
        lines = file.readlines()
    for (index, comment) in sorted(comment_map.items()):
        lines.insert(index, comment)
    with open(config_file, 'w') as file:
        file.write(''.join(lines))


def validate_config_against_schema(config: dict):
    """
    Validates config against JSON schema.
    Raises jsonschema.ValidationError if config is not valid.
    """
    schema_path = os.path.join(sourcedir, "UCC-UI-lib", "schema", "schema.json")
    with open(schema_path, "r") as f_schema:
        schema_raw = f_schema.read()
        schema = json.loads(schema_raw)
    return jsonschema.validate(instance=config, schema=schema)


def main():
    parser = argparse.ArgumentParser(description="Build the add-on")
    parser.add_argument(
        "--source",
        type=str,
        nargs='?',
        help="Folder containing the app.manifest and app source",
        default="package",
    )
    parser.add_argument(
        "--config",
        type=str,
        nargs='?',
        help="Path to configuration file, Defaults to GlobalConfig.json in parent directory of source provided",
        default=None
    )
    version = Version.from_git()
    if not version.stage:
        stage = 'R'
    else:
        stage = version.stage[:1]
    
    version_str = version.serialize(metadata=True,style=Style.SemVer)
    version_splunk = f"{version.base}{stage}{version.commit}"
    
    parser.add_argument(
        "--ta-version",
        type=str,
        help="Version of TA, Deafult version is version specified in the package such as app.manifest, app.conf, and globalConfig.json",
        default = version_splunk
    )
    args = parser.parse_args()
    ta_version = args.ta_version.strip()
    if not ta_version:
        ta_version = version_splunk
    if not os.path.exists(args.source):
        raise NotADirectoryError("{} not Found.".format(os.path.abspath(args.source)))

    # Setting default value to Config argument
    if not args.config:
        args.config = os.path.abspath(os.path.join(args.source, PARENT_DIR, "globalConfig.json"))

    clean_before_build()

    manifest= None
    with open(os.path.abspath(os.path.join(args.source, "app.manifest")), "r") as manifest_file:
        manifest = json.load(manifest_file)
        ta_name = manifest['info']['id']['name']

    if os.path.exists(args.config):
        try:
            with open(args.config, "r") as f_config:
                config_raw = f_config.read()
                config = json.loads(config_raw)
            validate_config_against_schema(config)
            logger.info("Config is valid")
        except jsonschema.ValidationError as e:
            logger.error("Config is not valid. Error: {}".format(e))
            sys.exit(1)

        update_ta_version(args.config, ta_version)

        # handle_update check schemaVersion and update globalConfig.json if required and return schema
        schema_content = handle_update(args.config)

        scheme = GlobalConfigBuilderSchema(schema_content, j2_env)
        
        ta_version = schema_content.get("meta").get("version")
        logger.info("Addon Version : " + ta_version)
        ta_tabs = schema_content.get("pages").get("configuration").get("tabs")
        ta_namespace = schema_content.get("meta").get("restRoot")
        import_declare_name = "import_declare_test"
        is_inputs = ("inputs" in schema_content.get("pages"))

        logger.info("Package ID is " + ta_name)

        logger.info("Copy UCC template directory")
        recursive_overwrite(
            os.path.join(sourcedir,"package"), os.path.join(outputdir, ta_name)
        )

        logger.info("Copy globalConfig to output")
        shutil.copyfile(
            args.config,
            os.path.join(outputdir, ta_name, "appserver", "static", "js", "build", "globalConfig.json"),
        )
        ucc_lib_target = os.path.join(outputdir, ta_name, "lib")
        logger.info(f"Install Addon Requirements into {ucc_lib_target} from {args.source}")
        install_libs(
            args.source ,
            ucc_lib_target
        )

        replace_token(args, ta_name)

        generate_rest(args, ta_name, scheme, import_declare_name)

        modify_and_replace_token_for_oauth_templates(
                args, ta_name, ta_tabs, schema_content.get('meta').get('version')
            )
        if is_inputs:
            add_modular_input(
                args, ta_name, schema_content, import_declare_name
            )
        else:
            handle_no_inputs(ta_name)
            
        make_modular_alerts(args, ta_name, ta_namespace, schema_content)

    else:
        logger.info("Addon Version : " + ta_version)
        logger.warning("Skipped installing UCC required python modules as GlobalConfig.json does not exist.")
        logger.warning("Skipped Generating UI components as GlobalConfig.json does not exist.")
        logger.info("Setting TA name as generic")

        ucc_lib_target = os.path.join(outputdir, ta_name, "lib")

        install_libs(
            args.source,
            ucc_lib_target=ucc_lib_target
        )

    ignore_list = get_ignore_list(ta_name, os.path.abspath(os.path.join(args.source, PARENT_DIR, ".uccignore")))
    remove_listed_files(ignore_list)
    copy_package_source(args, ta_name)

    #Update app.manifest
    with open(os.path.join(outputdir, ta_name,'VERSION'), 'w') as version_file:
        version_file.write(version_str)
        version_file.write("\n")
        version_file.write(ta_version)


    manifest= None
    with open(os.path.abspath(os.path.join(outputdir, ta_name, "app.manifest")), "r") as manifest_file:
        manifest = json.load(manifest_file)
        manifest['info']['id']['version'] = ta_version
    
    
    with open(os.path.abspath(os.path.join(outputdir, ta_name, "app.manifest")), "w") as manifest_file:
        manifest_file.write(json.dumps(manifest, indent=4, sort_keys=True))
        
    comment_map = save_comments(outputdir, ta_name)
    app_config = configparser.ConfigParser()        
    app_config.read_file(open(os.path.join(outputdir, ta_name,'default', "app.conf")))
    if not 'launcher' in app_config:
        app_config.add_section('launcher')
    if not 'id' in app_config:
        app_config.add_section('id')
    if not 'install' in app_config:
        app_config.add_section('install')
    if not 'package' in app_config:
        app_config.add_section('package')
    if not 'ui' in app_config:
        app_config.add_section('ui')

    app_config['launcher']['version'] = ta_version
    app_config['launcher']['description']=manifest['info']['description']
    
    app_config['id']['version'] = ta_version

    app_config['install']['build']=str(int(time.time()))
    app_config['package']['id']=manifest['info']['id']['name'] 

    app_config['ui']['label']=manifest['info']['title']

    with open(os.path.join(outputdir, ta_name,'default', "app.conf"), 'w') as configfile:
        app_config.write(configfile)
    #restore License header
    restore_comments(outputdir, ta_name, comment_map)
    
    #Copy Licenses
    license_dir = os.path.abspath(os.path.join(args.source, PARENT_DIR, "LICENSES"))
    
    if os.path.exists(license_dir):        
        logger.info("Copy LICENSES directory ")
        recursive_overwrite(license_dir, os.path.join(outputdir, ta_name,"LICENSES"))

    if os.path.exists(os.path.abspath(os.path.join(args.source,PARENT_DIR,"additional_packaging.py"))):
        sys.path.insert(0,os.path.abspath(os.path.join(args.source,PARENT_DIR)))
        from additional_packaging import additional_packaging
        additional_packaging(ta_name)


# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import json
import os
from jsl import AnyOfField, ArrayField, BooleanField, DictField, Document
from jsl import DocumentField, NumberField, OneOfField, StringField, UriField

###
# TODO Need to fix this: Make sure schemagenerator generates proper element for hook
# For time being do the following change in generated schema.json
# replace: "hook": {"$ref": "#/definitions/Hooks"} with "hook": {"type": "object"}
###

# Base Document Class with restricting properties population
class DocumentWithoutAddProp(Document):
    class Options(object):
        additional_properties = False


# Document Element with Label and Value element. Possible values are text, numeric and boolean types
class ValueLabelPair(DocumentWithoutAddProp):
    value = OneOfField([
        NumberField(),
        StringField(max_length=250),
        BooleanField()
    ])
    label = StringField(required=True, max_length=100)


class OAuthFields(DocumentWithoutAddProp):
    oauth_field = StringField(max_length=100)
    label = StringField(max_length=100)
    field = StringField(max_length=100)
    help = StringField(max_length=200)
    encrypted = BooleanField()



# Base Validator Wrapper component which is extension of DocumentWithoutAddProp Wrapper Component
class ValidatorBase(DocumentWithoutAddProp):
    errorMsg = StringField(max_length=400)


# MetaData component for detailing brief imformation of document/component
class Meta(DocumentWithoutAddProp):
    displayName = StringField(required=True, max_length=200)
    name = StringField(required=True, pattern="^[^<>\:\"\/\\\|\?\*]+$")
    restRoot = StringField(required=True, pattern="^\w+$")
    apiVersion = StringField(required=True, pattern="^(?:\d{1,3}\.){2}\d{1,3}$")
    version = StringField(required=True)
    schemaVersion = StringField( pattern="^(?:\d{1,3}\.){2}\d{1,3}$")


# Text validator for the String Field Value input
class StringValidator(ValidatorBase):
    type = StringField(required=True, enum=["string"])
    minLength = NumberField(required=True, minimum=0)
    maxLength = NumberField(required=True, minimum=0)


# Numeric validator for the Number Field Value input
class NumberValidator(ValidatorBase):
    type = StringField(required=True, enum=["number"])
    range = ArrayField(NumberField(), required=True)


# Regex validator for the text Field Value input
class RegexValidator(ValidatorBase):
    type = StringField(required=True, enum=["regex"])
    pattern = StringField(required=True)


# Email validator for the text Field Value input
class EmailValidator(ValidatorBase):
    type = StringField(required=True, enum=["email"])


# Ipv4 represenation validator
class Ipv4Validator(ValidatorBase):
    type = StringField(required=True, enum=["ipv4"])


# Date Validator
class DateValidator(ValidatorBase):
    type = StringField(required=True, enum=["date"])


# URL Validator
class UrlValidator(ValidatorBase):
    type = StringField(required=True, enum=["url"])

# Entity for Alert Actions
class AlertEntity(DocumentWithoutAddProp):
    field = StringField(required=True, pattern="^\w+$")
    label = StringField(required=True, max_length=30)
    type = StringField(required=True,
                       enum=["text", "singleSelect", "checkbox", "radio", "singleSelectSplunkSearch"])
    help = StringField(max_length=200)
    defaultValue = OneOfField([
        NumberField(),
        StringField(max_length=250),
        BooleanField()
    ])
    required = BooleanField()
    search = StringField(max_length=200)
    valueField = StringField(max_length=200)
    labelField = StringField(max_length=200)
    options = DictField(
        properties={
            "items": ArrayField(DocumentField(ValueLabelPair, as_ref=True))
        }
    )

##
#  Entity Form Field Component Wrapper having field name, label, field types, help/tooltips support, default value
#  Validators and UI Controls such as visibility etc
##
class Entity(DocumentWithoutAddProp):
    field = StringField(required=True, pattern="^\w+$")
    label = StringField(required=True, max_length=30)
    type = StringField(required=True,
                       enum=["custom", "text", "singleSelect", "checkbox", "multipleSelect", "radio", "placeholder", "oauth", "helpLink"])
    help = StringField(max_length=200)
    tooltip = StringField(max_length=250)
    defaultValue = OneOfField([
        NumberField(),
        StringField(max_length=250),
        BooleanField()
    ])
    options = DictField(
        properties={
            "disableSearch": BooleanField(),
            "autoCompleteFields": OneOfField([
                ArrayField(DictField(
                    properties={
                        "label": StringField(required=True, max_length=150),
                        "children": ArrayField(DocumentField(ValueLabelPair, as_ref=True), required=True)
                    }
                )),
                ArrayField(DocumentField(ValueLabelPair, as_ref=True))
            ]),
            "endpointUrl": StringField(max_length=350),
            "denyList": StringField(max_length=350),
            "allowList": StringField(max_length=350),
            "delimiter": StringField(max_length=1),
            "items": ArrayField(DocumentField(ValueLabelPair, as_ref=True)),
            "referenceName": StringField(max_length=250),
            "enable": BooleanField(),
            "placeholder": StringField(max_length=250),
            "display": BooleanField(),
            "labelField": StringField(max_length=250),
            "src": StringField(max_length=250),
            "defaultValue": StringField(max_length=250),
            "disableonEdit": BooleanField(),
            "basic": ArrayField(DocumentField(OAuthFields, as_ref=True)),
            "oauth": ArrayField(DocumentField(OAuthFields, as_ref=True)),
            "auth_type": ArrayField(StringField(max_length=100)),
            "auth_label": StringField(max_length=250),
            "oauth_popup_width": NumberField(),
            "oauth_popup_height": NumberField(),
            "oauth_timeout": NumberField(),
            "auth_code_endpoint": StringField(max_length=350),
            "access_token_endpoint": StringField(max_length=350),
            "text": StringField(max_length=50),
            "link": StringField()
        }
    )
    required = BooleanField()
    encrypted = BooleanField()
    # List of inbuilt field validator
    validators = ArrayField(AnyOfField([
        DocumentField(StringValidator, as_ref=True),
        DocumentField(NumberValidator, as_ref=True),
        DocumentField(RegexValidator, as_ref=True),
        DocumentField(EmailValidator, as_ref=True),
        DocumentField(Ipv4Validator, as_ref=True),
        DocumentField(UrlValidator, as_ref=True),
        DocumentField(DateValidator, as_ref=True)
    ]))


##
# Input Entity is super class of Entity to restrict the predefined field name holding entity field
##
class InputsEntity(Entity):
    # Prevnet Splunk reserved inputs field keys being used in the user customized inputs
    # https://jira.splunk.com/browse/ADDON-13014#comment-1493170

    field = StringField(
        required=True,
        pattern="(?!^(?:persistentQueueSize|queueSize|start_by_shell|output_mode|output_field|owner|app|sharing)$)(?:^\w+$)"
    )


##
# ConfigurationEntity is super class of Entity to restrict the predefined field name holding entity field such as
# output_mode
# output_field
# owner
# app
# sharing
# ##
class ConfigurationEntity(Entity):
    field = StringField(
        required=True,
        pattern="(?!^(?:output_mode|output_field|owner|app|sharing)$)(?:^\w+$)"
    )


##
# Table component schema with headers, additional info and customization row extension
##
class Table(DocumentWithoutAddProp):
    moreInfo = ArrayField(DictField(
        properties={
            "field": StringField(required=True, pattern="^\w+$"),
            "label": StringField(required=True, max_length=30),
            "mapping": DictField(required=False)
        }
    ))
    # Header field names needs to be display on UI
    header = ArrayField(DictField(
        properties={
            "field": StringField(required=True, pattern="^\w+$"),
            "label": StringField(required=True, max_length=30),
            "mapping": DictField(required=False),
            "customCell": DictField(required=False)
        }
    ), required=True)
    # custom Row implementation if required for special cases
    customRow = DictField(required=False)


##
# Input table having all functions of table and edit|delete|clone|enable/disable actions for each row
##
class InputsTable(Table):
    actions = ArrayField(StringField(enum=["edit", "delete", "clone", "enable"]), required=True)


##
# Configuration table having all functions of table and edit|delete|clone|enable/disable actions for each row
##
class ConfigurationTable(Table):
    actions = ArrayField(StringField(enum=["edit", "delete", "clone"]), required=True)


##
# Hook attribute scheme to define custom Hook for various events such as on load, on save and on save
##
class Hooks(DocumentWithoutAddProp):
    saveValidator = StringField(max_length=3000)
    onLoad = StringField(max_length=3000)
    onChange = StringField(max_length=3000)


##
# Tab Content holding the wrapper of Table and Rest Handler Mapping
##
class TabContent(DocumentWithoutAddProp):
    entity = ArrayField(DocumentField(ConfigurationEntity, as_ref=True), required=True)
    name = StringField(required=True, pattern="^[\/\w]+$", max_length=250)
    title = StringField(required=True, max_length=50)
    options = DocumentField(Hooks, as_ref=True)
    table = DocumentField(ConfigurationTable, as_ref=True)
    conf = StringField(required=False, max_length=100)
    restHandlerName = StringField(required=False, max_length=100)
    # Provisioning tab level hook on configuration page
    hook = DocumentField(Hooks, as_ref=True)


##
# ConfigurationPage Component having tabbing pages for configuration of various module
# Each tab is individual TabContent
##
class ConfigurationPage(DocumentWithoutAddProp):
    title = StringField(required=True, max_length=60)
    description = StringField(max_length=200)
    tabs = ArrayField(DocumentField(TabContent, as_ref=True), required=True, min_items=1)


##
# InputsPage Component having table and entity dialogue driven by service handler to add new entry
##
class InputsPage(DocumentWithoutAddProp):
    title = StringField(required=True, max_length=60)
    description = StringField(max_length=200)
    table = DocumentField(InputsTable, as_ref=True, required=True)
    services = ArrayField(DictField(
        properties={
            "name": StringField(required=True, pattern="^[0-9a-zA-Z][0-9a-zA-Z_-]*$", max_length=50),
            "title": StringField(required=True, max_length=100),
            "entity": ArrayField(DocumentField(InputsEntity, as_ref=True), required=True),
            "options": DocumentField(Hooks, as_ref=True),
            "groups": ArrayField(DictField(
                properties={
                    "options": DictField(
                        properties={
                            "isExpandable": BooleanField(),
                            "expand": BooleanField()
                        }
                    ),
                    "label": StringField(required=True, max_length=100),
                    "field": ArrayField(StringField(required=True, pattern="^\w+$"))
                }
            ), required=False),
            "style": StringField(required=False, enum=["page", "dialog"]),
            "hook": DictField(required=False),
            "conf": StringField(required=False, max_length=100),
            "restHandlerName": StringField(required=False, max_length=100)
        }
    ), required=True)
    menu = DictField(required=False)


##
# Main Page to holding configuration and input page
##
class Pages(DocumentWithoutAddProp):
    configuration = DocumentField(ConfigurationPage, as_ref=True, required=False)
    inputs = DocumentField(InputsPage, as_ref=True, required=False)


##
# Component holding Technology dict in active_response
##
class Technology(DocumentWithoutAddProp):
    version = ArrayField(StringField(required=True, pattern="^\d+(?:\.\d+)*$"),required=True, min_items=1)
    product = StringField(required=True, max_length=100)
    vendor = StringField(required=True, max_length=100)


##
# Main Component holding the alert actions 
##
class Alerts(DocumentWithoutAddProp):
    name = StringField(required=True, pattern="^[a-zA-Z0-9_]+$", max_length=100)
    label = StringField(required=True, max_length=100)
    description = StringField(required=True)
    activeResponse = DictField(properties={
                            "task": ArrayField(StringField(required=True), required=True, min_items=1),
                            "supportsAdhoc": BooleanField(required=True),
                            "subject": ArrayField(StringField(required=True), required=True, min_items=1),
                            "category": ArrayField(StringField(required=True), required=True, min_items=1),
                            "technology": ArrayField(DocumentField(Technology,as_ref=True),required=True, min_items=1),
                            "drilldownUri":StringField(required=False),
                            "sourcetype":StringField(required=False, pattern="^[a-zA-Z0-9:-_]+$", max_length=50)
                        }, required=False)
    entity = ArrayField(DocumentField(AlertEntity, as_ref=True))

##
# Root Component holding all pages and meta information
##
class UCCConfig(DocumentWithoutAddProp):
    meta = DocumentField(Meta, as_ref=True, required=True)
    pages = DocumentField(Pages, as_ref=True, required=True)
    alerts = ArrayField(DocumentField(Alerts, as_ref=True), required=False, min_items=1)

##
# SchemaGenerator responsible to generate schema json file holding information of Flow of UI based on UCCConfig Object
##
if __name__ == "__main__":
    formated = json.dumps(UCCConfig.get_schema(ordered=True), indent=4)
    formated = formated.replace("__main__.", "")

    cur_dir = os.path.dirname(__file__)
    with open(os.path.join(cur_dir, 'schema.json'), 'w+') as schema_handler:
        schema_handler.write(formated)


# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

# encode=utf-8

from builtins import range
from builtins import object
import time


class NumberMetricArregator(object):
    '''
    aggregate the metric with the the second interval.
    for minute and hour aggregate, use splunk to do it.
    '''
    def __init__(self, metric_name, event_writer, metric_tags=None, max_time_span=10):
        '''
        @params: metric_name - a literal string for the metric
        @params: an MetricEventWriter object which flushes the aggregated metric events
        @params: max_time_span - the aggregator creates max_time_span buckets.
                    Assume the metric event delay will not be large than max_time_span seconds.
        '''
        self.m_name = metric_name
        self.m_span = max_time_span
        self.event_writer = event_writer
        self.metric_tags = metric_tags
        self.oldest_metric_time = None
        self.buckets = dict()

    def aggregate_metric(self, record):
        '''
        this method is invoked in the worker thread.
        '''
        if self.oldest_metric_time is None:
            self.oldest_metric_time = record['ts']
        return self._aggregate_metric(record)

    def _aggregate_metric(self, record):
        raise NotImplemented('Children class must implement _aggregate_metric')

    def _format_metric(self, metric):
        raise NotImplemented('Children class must implement _format_metric')

    def get_formatted_metric(self, ts, metric):
        formatted_metric = self._format_metric(metric)
        formatted_metric['time_slot'] = ts
        formatted_metric['metric_name'] = self.m_name
        return formatted_metric

    def _flush_buckets(self, timestamp):
        if self.oldest_metric_time is None:
            return
        low_water_mark = timestamp - self.m_span
        if low_water_mark > self.oldest_metric_time:
            for ts in range(self.oldest_metric_time, low_water_mark):
                metric = self.buckets.get(ts, None)
                if metric:
                    self.event_writer.write_event(self.get_formatted_metric(ts, metric), self.metric_tags)
                    del self.buckets[ts]
            self.oldest_metric_time = low_water_mark

    def flush_all_event(self):
        for ts, metric in list(self.buckets.items()):
            self.event_writer.write_event(self.get_formatted_metric(ts, metric), self.metric_tags)


class NumberMetricSum(NumberMetricArregator):
    '''
    add all the numbers in the same time slot
    '''
    def __init__(self, metric_name, event_writer, metric_tags=None, max_time_span=10):
        super(NumberMetricSum, self).__init__(metric_name, event_writer, metric_tags, max_time_span)

    def _aggregate_metric(self, record):
        ts = record['ts']
        value = record['v']
        bucket_value = self.buckets.get(ts, None)
        if bucket_value is None:
            self.buckets[ts] = {'sum': value}
        else:
            bucket_value['sum'] = bucket_value['sum'] + value
            self.buckets[ts] = bucket_value
        self._flush_buckets(ts)

    def _format_metric(self, metric):
        return {'sum': metric['sum']}

class NumberMetricAverage(NumberMetricArregator):
    '''
    calculate the average number for the metric
    '''
    def __init__(self, metric_name, event_writer, metric_tags=None, max_time_span=10):
        super(NumberMetricAverage, self).__init__(metric_name, event_writer, metric_tags, max_time_span)

    def _format_metric(self, metric):
        return {'avg': float(metric['sum'])/metric['count'], 'max': metric['max'], 'min': metric['min']}

    def _aggregate_metric(self, record):
        ts = record['ts']
        value = record['v']
        bucket_value = self.buckets.get(ts, None)
        if bucket_value is None:
            self.buckets[ts] = {'sum': value, 'count': 1, 'max': value, 'min': value}
        else:
            self.buckets[ts] = {'sum': bucket_value['sum'] + value, 'count': bucket_value['count'] + 1, 'max': value if value > bucket_value['max'] else bucket_value['max'], 'min': value if value < bucket_value['min'] else bucket_value['min']}
        self._flush_buckets(ts)



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from builtins import zip
import json
import sys
import copy
import itertools

mapping_keys = {
    "activeResponse": "active_response",
    "supportsAdhoc": "supports_adhoc",
    "drilldownUri": "drilldown_uri",
    "entity": "parameters",
    "defaultValue": "default_value",
    "help": "help_string",
    "field": "name",
    "type": "format_type",
    "valueField": "value-field",
    "labelField": "label-field",
    "name": "short_name"
}

mapping_values = {
    "singleSelect": "dropdownlist",
    "singleSelectSplunkSearch": "dropdownlist_splunk_search"
}


def convert_list_to_dict(value_list):
    return_list = {}
    for each_dict in value_list:
        return_list[each_dict["label"]] = each_dict["value"]
    return return_list


def transform_params(parameter_list):
    for param in parameter_list:
        if param["format_type"] in ["dropdownlist", "radio"]:
            a = param.pop("options")
            param["possible_values"] = convert_list_to_dict(a["items"])
        elif param["format_type"] == "dropdownlist_splunk_search":
            value_field = param.pop("value-field")
            label_field = param.pop("label-field")
            search = param.pop("search")
            param["ctrl_props"] = {
                'value-field': value_field,
                'label-field': label_field,
                'search': search
            }


def iterdict(dictionary, result):
    '''
    This function replaces key and value with the ones required by add-on alert builder
    '''
    for key in dictionary:
        if key in mapping_keys:
            value = result.pop(key)
            if sys.version_info > (3, 0):
                result[mapping_keys[key]] =  value
            else:
                result[mapping_keys[key]] = value.encode(
                    'ascii', 'ignore') if isinstance(value, str) else value
            mapped_key = mapping_keys[key]
        else:
            mapped_key = key

        if isinstance(dictionary[key], dict):
            iterdict(dictionary[key], result[mapped_key])
        elif isinstance(dictionary[key], list):
            for dictionary_item, result_item in zip(dictionary[key], result[mapped_key]):
                if isinstance(dictionary_item, dict):
                    iterdict(dictionary_item, result_item)
        else:
            if result[mapped_key] in mapping_values:
                result[mapped_key] = mapping_values[result[mapped_key]]


def form_main_dict(alert, product_id, short_name):
    '''
    Forms the final dictionary required by add-on alert generator
    '''
    return {
        "product_id": product_id,
        "short_name": short_name,
        "global_settings": "",
        "html_setting": None,
        "build_components": {
            "conf": "True",
            "py": "True",
            "html": "True"
        },
        "schema.content": {
            "product_id": product_id,
            "short_name": short_name,
            "modular_alerts": alert
        }
    }


def normalize(schema_content, product_id, short_name):
    '''
    Process the globalConfig alert schema to generate structure required by add-on alert generator
    '''
    result = copy.deepcopy(schema_content)
    iterdict(schema_content, result)
    for alert in result["alerts"]:
        transform_params(alert["parameters"])
    return form_main_dict(result["alerts"], product_id, short_name)



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

# encoding = utf-8

from builtins import object
import time
import logging
import json

from solnlib import log

def message(app, current_time, event, tags):
    final_event = dict(event)
    final_event['mcollector_event_ts'] = current_time
    final_event['mcollector_target_app'] = app
    if tags:
        final_event['mcollector_tags'] = tags
    return json.dumps(final_event)


class MetricEventWriter(object):
    def __init__(self, app, config):
        '''
        config is a dict, which contains all the params for file writer
         - tag_black_list: a list of tags, if the tags are in the black list, do not write the event
         - tag_white_list: a list of tags, only write the event has this tags. white_list has higer priority than black_list
        '''
        assert app is not None
        assert isinstance(config, dict)
        self._app = app
        self._deny_list_tags = config.get('tag_black_list', [])
        self._allow_list_tags = config.get('tag_white_list', [])

    def write_event(self, ev, tags=[]):
        '''
        ev. event dict, it might be a hierarchy structure
        '''
        assert isinstance(ev, dict)
        assert isinstance(tags, list)
        if self._allow_list_tags:
            filter_tags = [t for t in tags if t in self._allow_list_tags]
            if filter_tags:
                self._flush_event(ev, tags)
            # else: no tags found in allow_list, just skip this event
            return

        if self._deny_list_tags and tags:
            filter_tags = [t for t in tags if t in self._deny_list_tags]
            if filter_tags:
                # skip this event
                return

        self._flush_event(ev, tags)

    def _flush_event(self, ev, tags):
        ctime = int(time.time() * 1000)  # ms
        self._flush_msg(message(self._app, ctime, ev, tags))

    def _flush_msg(self, msg):
        '''
        this should be implemented
        '''
        raise NotImplemented('_flush_msg should be implemented.')

    def update_config(self, config):
        self._deny_list_tags = config.get('tag_black_list', [])
        self._allow_list_tags = config.get('tag_white_list', [])


class FileEventWriter(MetricEventWriter):
    def __init__(self, app, config):
        super(FileEventWriter, self).__init__(app, config)
        self._logger_name = config.get('logger', 'default') + '_metric_events'
        self._logger = log.Logs().get_logger(self._logger_name)
        self._log_level = config.get('loglevel', 'INFO')
        log.Logs().set_level(self._log_level, self._logger_name)
        # reset the formater of log handler
        for handler in self._logger.handlers:
            handler.setFormatter(logging.Formatter('%(message)s'))

    def update_config(self, config):
        super(FileEventWriter, self).update_config(config)
        l_name = config.get('logger', 'default') + '_metric_events'
        if l_name != self._logger_name:
            self._logger_name = l_name
            self._logger = log.Logs().get_logger(self._logger_name)
        l_level = config.get('loglevel', 'INFO')
        if self._log_level != l_level:
            self._log_level = l_level
            log.Logs().set_level(self._log_level, self._logger_name)

    def _flush_msg(self, msg):
        self._logger.info(msg)

class SplunkStashFileWriter(MetricEventWriter):
    '''
    write a small file and use splunk rest upload this file
    '''
    def __init__(self, app, config):
        super(SplunkStashFileWriter, self).__init__(app, config)
        pass

    def update_config(self, config):
        super(FileEventWriter, self).update_config(config)
        pass

    def _flush_msg(self, msg):
        pass
